"""
.. _extra_index_url:

Add additional index URL to search
==================================

Pip accepts URLs for additional package indexes through ``--extra-index-url``.
The same option can be passed to ``pip-compile-multi`` during compilation.

This option is similar to `annotate_index`_.
The difference is that the passed URL is not saved in any of the files,
which is helpful if the URL contains private credentials.

.. code-block:: text

    --extra-index-url TEXT          Add additional package index URL to search
                                    for package versions. Can be supplied
                                    multiple times.


.. warning::

    Using ``--extra-index-url`` option to search for packages that are not in
    the main repository (such as private packages) is unsafe, per a security vulnerability
    called `dependency confusion`_: an attacker can claim the package on
    the public repository in a way that will ensure it gets chosen over the private package.

    Use the ``--index-url`` option in pip’s configuration file or command line
    to specify the feed, overriding the default.
    Avoid the ``--extra-index-url`` option, which is additive and may lead
    to having multiple indexes.

    See `pip documentation`_ for details.

.. _pip documentation: https://pip.pypa.io/en/stable/cli/pip_install/#examples
.. _dependency confusion: https://azure.microsoft.com/en-us/resources/\
        3-ways-to-mitigate-risk-using-private-package-feeds/
"""

from .base import BaseFeature, ClickOption


class ExtraIndexUrl(BaseFeature):
    """Forward extra index URLs to to pip-compile."""

    _OPTION = '--extra-index-url'
    OPTION_NAME = 'extra_index_url'
    CLICK_OPTION = ClickOption(
        long_option=_OPTION,
        multiple=True,
        help_text=(
            'Add additional package index URL to search for package versions. '
            'Can be supplied multiple times.'
        )
    )

    def pin_options(self):
        """Pin command options."""
        if self.value:
            parts = []
            for url in self.value:
                parts.extend([self._OPTION, url])
            return parts
        return []



"""Human-friendly interface to pip-compile-multi"""
import functools
import logging

import click

from .options import OPTIONS
from .config import read_config, read_sections
from .actions import recompile
from .verify import verify_environments


logger = logging.getLogger("pip-compile-multi")


@click.group()
def cli():
    """Human-friendly interface to pip-compile-multi"""
    logging.basicConfig(level=logging.DEBUG, format="%(message)s")


@cli.command()
def lock():
    """Lock new dependencies without upgrading"""
    OPTIONS['upgrade'] = False
    run_configurations(recompile, read_config)


@cli.command()
def upgrade():
    """Upgrade locked dependency versions"""
    OPTIONS['upgrade'] = True
    OPTIONS['upgrade_packages'] = []
    run_configurations(recompile, read_config)


@cli.command()
@click.pass_context
def verify(ctx):
    """Upgrade locked dependency versions"""
    oks = run_configurations(
        skipper(verify_environments),
        read_sections,
    )
    ctx.exit(0
             if False not in oks
             else 1)


def skipper(func):
    """Decorator that memorizes base_dir, in_ext and out_ext from OPTIONS
    and skips execution for duplicates.
    """
    @functools.wraps(func)
    def wrapped():
        """Dummy docstring to make pylint happy."""
        key = (OPTIONS['base_dir'], OPTIONS['in_ext'], OPTIONS['out_ext'])
        if key not in seen:
            seen[key] = func()
        return seen[key]
    seen = {}
    return wrapped


def run_configurations(callback, sections_reader):
    """Parse configurations and execute callback for matching."""
    base = {
        'base_dir': 'requirements',
        'in_ext': 'in',
        'out_ext': 'txt',
        'autoresolve': True,
    }
    sections = sections_reader()
    if sections is None:
        logger.info("Configuration not found in .ini files. "
                    "Running with default settings")
        recompile()
    elif sections == []:
        logger.info("Configuration does not match current runtime. "
                    "Exiting")
    results = []
    for section, options in sections:
        OPTIONS.clear()
        OPTIONS.update(base)
        OPTIONS.update(options)
        logger.debug("Running configuration from section \"%s\". OPTIONS: %r",
                     section, OPTIONS)
        results.append(callback())
    return results


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



"""
Use Cache
=========

By default ``pip-compile-multi`` executes ``pip-compile`` with
``--rebuild`` flag.
This flag clears any caches upfront and rebuilds from scratch.
Such a strategy has proven to be more reliable in `edge cases`_,
but causes significant performance degradation.

Option ``--use-cache`` removes ``--rebuild`` flag from the call
to ``pip-compile``.

.. code-block:: text

    -u, --use-cache             Use pip-tools cache to speed up compilation.

.. note::

    When using ``--use-cache``, ``pip-compile-multi`` can
    run **10 times faster**. But if you run into "magical" issues,
    try to rerun compilation without this flag first.

.. _edge cases: https://github.com/jazzband/pip-tools/issues?q=--rebuild
"""

from .base import ClickOption
from .forward import ForwardOption


class UseCache(ForwardOption):
    """Use pip-tools cache, or rebuild from scratch."""

    OPTION_NAME = 'use_cache'
    CLICK_OPTION = ClickOption(
        long_option='--use-cache',
        short_option='-u',
        is_flag=True,
        default=False,
        help_text='Use pip-tools cache to speed up compilation.',
    )
    disabled_pin_options = ['--rebuild']



"""
.. _autoresolve:

Autoresolve cross-file conflicts
================================

Compile requirements file, that references all other files first,
and than use it as a constraint.

.. code-block:: text

    --autoresolve/--no-autoresolve Automatically resolve
                                   cross-file conflicts.

This strategy allows to resolve cross-file conflicts of two types:

1. Files FOO and BAR both have dependency PKG, but compile it to different versions.
2. FOO has PKG resolved to version 3. BAR references FOO, but also
   has another dependency, that constraints PKG to version 2.

By compiling all-including file (aka *sink*), that references both FOO and BAR first,
pip-compile-multi generates conflict-free set of versions.

After that, this compiled file is passed as a constraint for compiling
all requirements files.

As the last step, the *sink* file is compiled again preserving reference files
and skipping duplicate packages.

.. note::

    This feature works only if your project has a single requirements file,
    that references (directly or indirectly) all other files.
"""

from pipcompilemulti.utils import recursive_refs
from .base import BaseFeature, ClickOption


class Autoresolve(BaseFeature):
    """Detect sink file and use it unless the feature is explicitly disabled."""

    OPTION_NAME = 'autoresolve'
    CLICK_OPTION = ClickOption(
        long_option='--autoresolve/--no-autoresolve',
        is_flag=True,
        default=False,
        help_text='Automatically resolve cross-file conflicts.',
    )

    def __init__(self):
        self._sink_path = None

    @property
    def enabled(self):
        """Whether feature was explicitly disabled or not."""
        return self.value

    def on_discover(self, env_confs):
        """Save set of all (recursive) included environments."""
        self._sink_path = self._find_sink(env_confs)

    def sink_path(self):
        """Return sink path if it's enabled. Otherwise None"""
        return self._sink_path if self.enabled else None

    @staticmethod
    def _find_sink(envs):
        """Try to find requirements sink.

        Sink is a requirements file that references all other
        requirement files.

        If no sink exists, return None.

        >>> Autoresolve._find_sink([
        ...  {'in_path': 'base', 'refs': set()},
        ...  {'in_path': 'test', 'refs': {'base'}},
        ...  {'in_path': 'local', 'refs': {'test', 'base'}},
        ... ])
        'local'
        >>> Autoresolve._find_sink([
        ...  {'in_path': 'base', 'refs': set()},
        ...  {'in_path': 'test', 'refs': {'base'}},
        ...  {'in_path': 'doc', 'refs': set()},
        ... ])
        >>> Autoresolve._find_sink([
        ...  {'in_path': 'base', 'refs': set()},
        ...  {'in_path': 'foo', 'refs': {'base'}},
        ...  {'in_path': 'bar', 'refs': {'base'}},
        ...  {'in_path': 'all', 'refs': {'foo', 'bar'}},
        ... ])
        'all'
        """
        all_envs = {env['in_path'] for env in envs}
        for env in envs:
            included_envs = set(recursive_refs(envs, env['in_path'])) | {env['in_path']}
            if all_envs == included_envs:
                return env['in_path']
        return None



"""
.. _annotate_index:

Add index URL annotation
========================

Control addition of ``--index-url`` options to the output files.
Corresponds to ``pip-compile``'s  ``--emit-index-url / --no-emit-index-url`` flag.

The URL can be defined in the input file as in this example:

.. code-block:: text

    --index-url=https://pypi.tuna.tsinghua.edu.cn/simple/
    six
    click

This option is similar to `extra_index_url`_.
The difference is that URL is saved in the output files,
which means that it doesn't need to be configured in every environment.

.. code-block:: text

    --annotate-index / --no-annotate-index
                                    Add index URL to generated files
                                    (default false)

Note: the default behavior is not to add the index, i.e., ``--no-annotate-index``.
"""

from .base import ClickOption
from .forward import ForwardOption


class AnnotateIndex(ForwardOption):
    """Optionally annotate the index URL to the generated files."""

    OPTION_NAME = "annotate_index"
    CLICK_OPTION = ClickOption(
        long_option="--annotate-index/--no-annotate-index",
        default=False,
        is_flag=True,
        help_text="Add the index URL to generated files (default false).",
    )
    enabled_pin_options = ["--emit-index-url"]
    disabled_pin_options = ["--no-emit-index-url"]



"""
Compatible Releases
===================

`PEP-440`_ describes compatible release operator ``~=``.
Sometimes it's useful to have some of the dependencies
pinned using this operator.
For example, rapidly changing internal libraries.
The format for this option is

.. code-block:: text

    -c, --compatible TEXT

where TEXT is a `glob`_ pattern for library name.
This option can be supplied multiple times.

.. _glob: https://en.wikipedia.org/wiki/Glob_(programming)
.. _PEP-440: https://www.python.org/dev/peps/pep-0440/#compatible-release
"""

from fnmatch import fnmatch

from .base import BaseFeature, ClickOption


class Compatible(BaseFeature):
    """Use ~= for selected packages."""

    OPTION_NAME = 'compatible_patterns'
    CLICK_OPTION = ClickOption(
        long_option='--compatible',
        short_option='-c',
        multiple=True,
        help_text='Glob expression for packages with compatible (~=) '
                  'version constraint. Can be supplied multiple times.'
    )

    @property
    def patterns(self):
        """Use empty list as the default."""
        return self.value or []

    def constraint(self, package_name):
        """Return ``~=`` if package_name matches patterns, ``==`` otherwise.

        >>> from pipcompilemulti.options import OPTIONS
        >>> feature = Compatible()
        >>> OPTIONS[feature.OPTION_NAME] = ['xxx']
        >>> feature.constraint('package')
        '=='
        >>> feature.constraint('xxx')
        '~='
        """
        return '~=' if self.is_matched(package_name) else '=='

    def is_matched(self, package_name):
        """Whether package name matches one of configured glob patterns."""
        package = package_name.lower()
        for pattern in self.patterns:
            if fnmatch(package, pattern):
                return True
        return False



"""
Backtracking resolver
=====================

Pip has an option to enable `backtracking`_ conflict resolution logic,
which can automatically downgrade some dependencies to meet constraints
from other packages.
See also the `a note on resolvers`_ in ``pip-compile`` project.

.. code-block:: text

    --backtracking / --no-backtracking
                                Enable backtracking resolver. Translates to
                                pip-compile --resolver=backtracking option.

.. _backtracking: https://pip.pypa.io/en/latest/user_guide/\
        #changes-to-the-pip-dependency-resolver-in-20-3-2020
.. _a note on resolvers: https://github.com/jazzband/pip-tools#a-note-on-resolvers
"""

from .base import ClickOption
from .forward import ForwardOption


class Backtracking(ForwardOption):
    """Enable backtracking resolver."""

    OPTION_NAME = 'backtracking'
    CLICK_OPTION = ClickOption(
        long_option='--backtracking/--no-backtracking',
        is_flag=True,
        default=False,
        help_text='Enable backtracking resolver. Translates to '
                  'pip-compile --resolver=backtracking option.'
    )
    enabled_pin_options = ['--resolver=backtracking']
    disabled_pin_options = ['--resolver=legacy']



"""Environment class"""

import os
import re
import sys
import logging
import subprocess

from .dependency import Dependency
from .features import FEATURES
from .deduplicate import PackageDeduplicator
from .utils import extract_env_name


logger = logging.getLogger("pip-compile-multi")


class Environment(object):
    """requirements file"""

    RE_REF = re.compile(r'^(?:-r|--requirement)\s*(?P<path>\S+).*$')
    RE_COMMENT = re.compile(r'^\s*#.*$')

    def __init__(self, in_path, deduplicator=None):
        """
        Args:
            in_path: relative path to input file, e.g. requirements/base.in
        """
        self.in_path = in_path
        self._dedup = deduplicator or PackageDeduplicator()
        self.ignore = self._dedup.ignored_packages(in_path)
        self.packages = {}
        self._outfile_pkg_names = None

    def maybe_create_lockfile(self):
        """
        Write recursive dependencies list to outfile unless the goal is
        to upgrade specific package(s) which don't already appear.
        Populate package ignore set in either case and return
        boolean indicating whether outfile was written.
        """
        logger.info(
            "Locking %s to %s. References: %r",
            self.infile,
            self.outfile,
            sorted(self._dedup.recursive_refs(self.in_path)),
        )
        if not FEATURES.affected(self.in_path):
            self.fix_lockfile()  # populate ignore set
            return False
        self.create_lockfile()
        return True

    def create_lockfile(self):
        """
        Write recursive dependencies list to outfile
        with hard-pinned versions.
        Then fix it.
        """
        original_in_file = ""
        sink_out_path = FEATURES.sink_out_path()
        try:
            if sink_out_path and sink_out_path != self.outfile:
                original_in_file = self._read_infile()
                self._inject_sink()
            with subprocess.Popen(self.pin_command, **FEATURES.pipe_arguments()) as process:
                stdout, stderr = process.communicate()
        finally:
            if original_in_file:
                self._restore_in_file(original_in_file)
        if process.returncode == 0:
            self.fix_lockfile()
        else:
            logger.critical("ERROR executing %s", ' '.join(self.pin_command))
            logger.critical("Exit code: %s", process.returncode)
            if stdout:
                logger.critical(stdout.decode('utf-8'))
            if stderr:
                logger.critical(stderr.decode('utf-8'))
            raise RuntimeError("Failed to pip-compile {0}".format(self.infile))

    @classmethod
    def parse_references(cls, filename):
        """
        Read filename line by line searching for pattern:

        -r file.in
        or
        --requirement file.in

        return set of matched file names.
        E.g. {'file1.in', 'file2.in'}
        """
        references = set()
        with open(filename, encoding="utf-8") as fobj:
            for line in fobj:
                matched = cls.RE_REF.match(line)
                if matched:
                    references.add(matched.group('path'))
        return references

    @property
    def name(self):
        """Generate environment name from in_path."""
        return extract_env_name(self.in_path)

    @property
    def infile(self):
        """Path of the input file"""
        return self.in_path

    @property
    def outfile(self):
        """Path of the output file"""
        return FEATURES.compose_output_file_path(self.in_path)

    @property
    def pin_command(self):
        """Compose pip-compile shell command"""
        # Use the same interpreter binary
        python = sys.executable or 'python'
        parts = [
            python, '-m', 'piptools', 'compile',
            '--no-header',
            '--verbose',
        ]
        parts.extend(FEATURES.pin_options(self.in_path))
        parts.extend(['--output-file', self.outfile, self.infile])
        return parts

    def fix_lockfile(self):
        """Run each section of outfile through fix_pin"""
        with open(self.outfile, 'rt', encoding="utf-8") as fp:
            sections = [
                self.fix_pin(section)
                for section in self.parse_sections(self.concatenated(fp))
            ]
        with open(self.outfile, 'wt', encoding="utf-8") as fp:
            fp.writelines([
                section + '\n'
                for section in sections
                if section is not None
            ])
        self._dedup.register_packages_for_env(self.in_path, self.packages)

    @staticmethod
    def concatenated(fp):
        r"""Read lines from fp concatenating on backslash (\\)

        >>> env = Environment('')
        >>> list(env.concatenated([
        ...     'pkg', 'pkg  # comment', 'pkg', '# comment', '# one more',
        ...     'foo', '  # via', '', '  # pkg',
        ... ]))
        ['pkg', 'pkg  # comment', 'pkg', '# comment', '# one more', 'foo', '  # via', '', '  # pkg']
        """
        line_parts = []
        for line in fp:
            line = line.rstrip()
            if line.endswith('\\'):
                line_parts.append(line[:-1].rstrip())
            else:
                line_parts.append(line)
                yield ' '.join(line_parts)
                line_parts[:] = []
        if line_parts:
            # Impossible:
            raise RuntimeError("Compiled file ends with backslash \\")

    def parse_sections(self, lines):
        r"""Combine lines with following comments into sections.

        >>> env = Environment('')
        >>> list(env.parse_sections([
        ...     'pkg', 'pkg  # comment', 'pkg', '# comment', '# one more',
        ...     'foo', '  # via', '', '  # pkg',
        ... ]))
        ['pkg', 'pkg  # comment', 'pkg\n# comment\n# one more', 'foo\n  # via', '\n  # pkg']
        """
        section = []
        for line in lines:
            if self.RE_COMMENT.match(line):
                section.append(line)
            else:
                if section:
                    yield '\n'.join(section)
                section = [line]
        if section:
            yield '\n'.join(section)

    def fix_pin(self, section):
        """
        Fix dependency by removing post-releases from versions
        and loosing constraints on internal packages.
        Drop packages from ignore set

        Also populate packages set
        """
        dep = Dependency(section)
        if dep.valid:
            if dep.package in self.ignore:
                ignored_version = self.ignore[dep.package]
                if ignored_version is not None:
                    # ignored_version can be None to disable conflict detection
                    if dep.version and dep.version != ignored_version:
                        logger.error(
                            "Package %s was resolved to different "
                            "versions in different environments: %s and %s",
                            dep.package, dep.version, ignored_version,
                        )
                        raise RuntimeError(
                            "Please add constraints for the package "
                            "version listed above"
                        )
                return None
            self.packages[dep.package] = dep.version
            dep.drop_post(self.in_path)
            return dep.serialize()
        return section.rstrip()

    def add_references(self, other_in_paths):
        """Add references to other_in_paths in outfile"""
        if not other_in_paths:
            # Skip on empty list
            return
        with open(self.outfile, 'rt', encoding="utf-8") as fp:
            header, body = self.split_header(fp)
        with open(self.outfile, 'wt', encoding="utf-8") as fp:
            fp.writelines(header)
            fp.writelines(
                '-r {0}\n'.format(
                    FEATURES.compose_output_file_path(other_in_path)
                )
                for other_in_path in sorted(other_in_paths)
            )
            fp.writelines(body)

    @staticmethod
    def split_header(fp):
        """
        Read file pointer and return pair of lines lists:
        first - header, second - the rest.
        """
        body_start, header_ended = 0, False
        lines = []
        for line in fp:
            if line.startswith('#') and not header_ended:
                # Header text
                body_start += 1
            else:
                header_ended = True
            lines.append(line)
        return lines[:body_start], lines[body_start:]

    def replace_header(self, header_text):
        """Replace pip-compile header with custom text"""
        with open(self.outfile, 'rt', encoding="utf-8") as fp:
            _, body = self.split_header(fp)
        with open(self.outfile, 'wt', encoding="utf-8") as fp:
            fp.write(header_text)
            fp.writelines(body)

    def _read_infile(self):
        with open(self.infile, "rt", encoding="utf-8") as fp:
            return fp.read()

    def _restore_in_file(self, content):
        with open(self.infile, "wt", encoding="utf-8") as fp:
            return fp.write(content)

    def _inject_sink(self):
        rel_sink_out_path = os.path.normpath(os.path.relpath(
            FEATURES.sink_out_path(),
            os.path.dirname(self.infile),
        ))
        with open(self.infile, "at", encoding="utf-8") as fp:
            return fp.write("\n\n-c {}\n".format(rel_sink_out_path))



"""Aggregate all features in a single controller."""

import os
from functools import wraps

from .add_hashes import AddHashes
from .annotate_index import AnnotateIndex
from .autoresolve import Autoresolve
from .backtracking import Backtracking
from .base_dir import BaseDir
from .build_isolation import BuildIsolation
from .compatible import Compatible
from .emit_trusted_host import EmitTrustedHost
from .extra_index_url import ExtraIndexUrl
from .file_extensions import InputExtension, OutputExtension
from .forbid_post import ForbidPost
from .header import CustomHeader
from .limit_envs import LimitEnvs
from .limit_in_paths import LimitInPaths
from .live_output import LiveOutput
from .skip_constraint_comments import SkipConstraintComments
from .unsafe import AllowUnsafe
from .upgrade import UpgradeAll, UpgradeSelected
from .use_cache import UseCache


class FeaturesController:
    """Gateway to a list of features."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        self.add_hashes = AddHashes(self)
        self.allow_unsafe = AllowUnsafe()
        self.annotate_index = AnnotateIndex()
        self.autoresolve = Autoresolve()
        self.backtracking = Backtracking()
        self.base_dir = BaseDir()
        self.build_isolation = BuildIsolation()
        self.compatible = Compatible()
        self.emit_trusted_host = EmitTrustedHost()
        self.extra_index_url = ExtraIndexUrl()
        self.forbid_post = ForbidPost()
        self.header = CustomHeader()
        self.input_extension = InputExtension()
        self.limit_envs = LimitEnvs(self)
        self.limit_in_paths = LimitInPaths()
        self.live_output = LiveOutput()
        self.output_extension = OutputExtension()
        self.skip_constraint_comments = SkipConstraintComments()
        self.upgrade_all = UpgradeAll(self)
        self.upgrade_selected = UpgradeSelected(self)
        self.use_cache = UseCache()
        self._features = [
            self.add_hashes,
            self.allow_unsafe,
            self.annotate_index,
            self.autoresolve,
            self.backtracking,
            self.base_dir,
            self.build_isolation,
            self.compatible,
            self.emit_trusted_host,
            self.extra_index_url,
            self.forbid_post,
            self.header,
            self.input_extension,
            self.limit_envs,
            self.limit_in_paths,
            self.live_output,
            self.output_extension,
            self.skip_constraint_comments,
            self.upgrade_all,
            self.upgrade_selected,
            self.use_cache,
        ]

    def bind(self, command):
        """Bind all features to click command."""
        @wraps(command)
        def save_command_options(*args, **kwargs):
            """Save option values and call original command without it."""
            for feature in self._features:
                feature.extract_option(kwargs)
            return command(*args, **kwargs)

        for feature in self._features:
            save_command_options = feature.bind(save_command_options)
        return save_command_options

    def pin_options(self, in_path):
        """Return list of options to pin command."""
        options = []
        options.extend(self.add_hashes.pin_options(in_path))
        options.extend(self.allow_unsafe.pin_options())
        options.extend(self.annotate_index.pin_options())
        options.extend(self.backtracking.pin_options())
        options.extend(self.build_isolation.pin_options())
        options.extend(self.emit_trusted_host.pin_options())
        options.extend(self.extra_index_url.pin_options())
        options.extend(self.upgrade_all.pin_options())
        options.extend(self.upgrade_selected.pin_options())
        options.extend(self.use_cache.pin_options())
        return options

    def compose_input_file_path(self, basename):
        """Return input file path by environment name."""
        return self.base_dir.file_path(
            self.input_extension.compose_input_file_name(basename)
        )

    def compose_output_file_path(self, in_path):
        """Return output file path by environment name."""
        return self.output_extension.compose_output_file_path(in_path)

    def drop_post(self, in_path, package_name, version):
        """Whether post versions are forbidden for passed environment name."""
        if self.forbid_post.post_forbidden(in_path):
            return self.forbid_post.drop_post(version)
        if self.compatible.is_matched(package_name):
            return self.forbid_post.drop_post(version)
        return version

    def constraint(self, package_name):
        """Return ``~=`` if package_name matches patterns, ``==`` otherwise."""
        return self.compatible.constraint(package_name)

    def on_discover(self, env_confs):
        """Configure features with a list of discovered environments.

        Returns a new possibly shorter env list.
        """
        self.upgrade_selected.reset()
        self.limit_envs.on_discover(env_confs)
        self.limit_in_paths.on_discover(env_confs)
        limited_env_confs = [env for env in env_confs if self.included(env['in_path'])]
        self.add_hashes.on_discover(limited_env_confs)
        self.autoresolve.on_discover(limited_env_confs)
        return limited_env_confs

    def affected(self, in_path):
        """Whether environment is affected by upgrade command."""
        if self.upgrade_all.enabled:
            return True
        if self.upgrade_selected.affected(in_path):
            return True
        return in_path == self.autoresolve.sink_path()

    def included(self, in_path):
        """Whether in_path is included directly or by reference."""
        return (
            self.limit_envs.included(in_path) and self.limit_in_paths.included(in_path)
        )

    def get_header_text(self):
        """Text to put in the beginning of each generated file."""
        return self.header.text

    def sink_in_path(self):
        """Return input sink path if it's enabled. Otherwise None"""
        return self.autoresolve.sink_path()

    def sink_out_path(self):
        """Return sink output path if it's enabled and exists. Otherwise None"""
        infile = self.autoresolve.sink_path()
        if not infile:
            return None
        outfile = self.compose_output_file_path(infile)
        if not os.path.exists(outfile):
            return None
        return outfile

    def process_dependency_comments(self, comment):
        """Process comments of locked dependency (e.g. # via xxx)."""
        return self.skip_constraint_comments.process_dependency_comments(comment)

    def pipe_arguments(self):
        """Values for stdout and stderr arguments to subprocess.Popen."""
        return self.live_output.pipe_arguments()



"""
Requirements Files Extensions
=============================

By default ``pip-compile-multi`` compiles ``*.txt`` from ``*.in`` files.
While this is a common naming pattern, each project can use it's own:

.. code-block:: text

    -i, --in-ext TEXT      File extension of input files
    -o, --out-ext TEXT     File extension of output files
"""

import os
from .base import BaseFeature, ClickOption


class InputExtension(BaseFeature):
    """Override input file extension."""

    OPTION_NAME = 'in_ext'
    CLICK_OPTION = ClickOption(
        long_option='--in-ext',
        short_option='-i',
        default="in",
        is_flag=False,
        help_text='File extension of input files.',
    )

    def compose_input_file_name(self, base_name):
        """Compose file name given environment name.

        >>> InputExtension().compose_input_file_name('base')
        'base.in'
        """
        return '{0}.{1}'.format(base_name, self.value)


class OutputExtension(BaseFeature):
    """Override output file extension."""

    OPTION_NAME = 'out_ext'
    CLICK_OPTION = ClickOption(
        long_option='--out-ext',
        short_option='-o',
        default="txt",
        is_flag=False,
        help_text='File extension of output files.',
    )

    def compose_output_file_path(self, in_path):
        """Compose file name given environment name.

        >>> OutputExtension().compose_output_file_path('sub/base.in')
        'sub/base.txt'
        """
        return '{0}.{1}'.format(os.path.splitext(in_path)[0], self.value)



"""
Limit environments
==================

.. warning::

    This flag is deprecated and will be removed in future releases.
    Use :ref:`limit-in-files` instead.


By default ``pip-compile-multi`` compiles all ``.in`` files in ``requirements`` directory.
To limit compilation to only a subset, use

.. code-block:: text

    -n, --only-name TEXT        Compile only for passed environment names and
                                their references. Can be supplied multiple
                                times.

For example, to compile one file under Python2.7 and another under Python3.6, run:

.. code-block:: text

    $ virtual-env27/bin/pip-compile-multi -n deps27
    Locking requirements/deps27.in to requirements/deps27.txt. References: []
    $ virtual-env36/bin/pip-compile-multi -n deps36
    Locking requirements/deps36.in to requirements/deps36.txt. References: []
"""

from .base import ClickOption
from .limit_in_paths import LimitInPaths


class LimitEnvs(LimitInPaths):
    """Limit discovered environments to specified subset."""

    OPTION_NAME = 'include_names'
    CLICK_OPTION = ClickOption(
        long_option='--only-name',
        short_option='-n',
        multiple=True,
        help_text='Compile only for passed environment names and their '
                  'references. Can be supplied multiple times.',
    )

    def __init__(self, controller):
        # pylint: disable=super-with-arguments
        self._controller = controller
        super(LimitEnvs, self).__init__()

    @property
    def direct_envs(self):
        """Set of environments included by command line options."""
        return set(
            self._controller.compose_input_file_path(env_name)
            for env_name in self.value or []
        )



"""Common functionality for features activated by command line option."""

import click

from ..options import OPTIONS


class ClickOption:
    """Click option parameters."""

    def __init__(self,
                 long_option='',
                 short_option='',
                 is_flag=False,
                 default=None,
                 multiple=False,
                 help_text=''):
        self.long_option = long_option
        self.short_option = short_option
        self.is_flag = is_flag
        self.default = default
        self.multiple = multiple
        self.help_text = help_text

    def decorate(self, command):
        """Decorate click command with this option."""
        return self.decorator()(command)

    def decorator(self):
        """Create click command decorator with this option."""
        args = [self.long_option]
        kwargs = {
            "is_flag": self.is_flag,
            "multiple": self.multiple,
            "help": self.help_text,
        }
        if self.short_option:
            args.append(self.short_option)
        if self.default:
            kwargs.update(default=self.default)
        return click.option(*args, **kwargs)

    @property
    def argument_name(self):
        """Generate command argument name from long option.

        >>> ClickOption("--param-name").argument_name
        'param_name'
        >>> ClickOption("--param-name/--no-param-name").argument_name
        'param_name'
        """
        return self.long_option.lstrip('--').split('/', 1)[0].replace('-', '_')


class BaseFeature:
    """Base class for features."""

    OPTION_NAME = None
    CLICK_OPTION = None

    def bind(self, command):
        """Bind feature's click option to passed command."""
        return self.CLICK_OPTION.decorate(command)

    def extract_option(self, kwargs):
        """Pop option value from kwargs and save it in OPTIONS.

        If option was saved before and new value is the same as default,
        then keep previous value.
        This allows passing options both before and after ``verify`` command.
        """
        new_value = kwargs.pop(self.CLICK_OPTION.argument_name)
        if self.OPTION_NAME in OPTIONS and new_value == self.CLICK_OPTION.default:
            # Do not overwrite with default if already set.
            return
        OPTIONS[self.OPTION_NAME] = new_value

    @property
    def value(self):
        """Option value."""
        return OPTIONS.get(self.OPTION_NAME, self.CLICK_OPTION.default)

    @value.setter
    def value(self, new_value):
        OPTIONS[self.OPTION_NAME] = new_value



"""The current stable version of command line interface."""

import os
import sys
import logging
from traceback import print_exception

import click

from .actions import recompile
from .verify import verify_environments
from .features import FEATURES


THIS_FILE = os.path.abspath(__file__)


@click.group(invoke_without_command=True)
@click.pass_context
@FEATURES.bind
def cli(ctx):
    """Recompile"""
    logging.basicConfig(level=logging.DEBUG, format="%(message)s")
    sys.excepthook = exception_hook
    if ctx.invoked_subcommand is None:
        recompile()


@cli.command()
@click.pass_context
@FEATURES.bind
def verify(ctx):
    """
    For each environment verify hash comments and report failures.
    If any failure occured, exit with code 1.
    """
    sys.excepthook = exception_hook
    ctx.exit(0
             if verify_environments()
             else 1)


def exception_hook(exctype, value, traceback):
    """Strip exception printout above this module."""
    print_exception(exctype, value, trim_traceback(traceback))


def trim_traceback(traceback):
    """Trim traceback top so it starts with this module.

    Return original traceback if this module is not found.
    """
    level = 0
    new_traceback = traceback
    while new_traceback is not None:
        file_path = new_traceback.tb_frame.f_code.co_filename
        if THIS_FILE.startswith(file_path):
            return new_traceback
        level += 1
        new_traceback = new_traceback.tb_next
    return traceback


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python
"""High level actions to be called from CLI"""

import logging

from .discover import discover
from .environment import Environment
from .verify import generate_robust_hash_comment
from .features import FEATURES
from .deduplicate import PackageDeduplicator


logger = logging.getLogger("pip-compile-multi")


def recompile():
    """Compile requirements files for all environments."""
    env_confs = FEATURES.on_discover(
        discover(FEATURES.compose_input_file_path('*'))
    )
    deduplicator = PackageDeduplicator()
    deduplicator.on_discover(env_confs)
    sink_in_path = FEATURES.sink_in_path()
    if sink_in_path:
        sink_env = Environment(in_path=sink_in_path)
        logger.info(
            "Creating a temporary file with all dependencies at %s",
            sink_env.outfile,
        )
        sink_env.create_lockfile()
    compile_topologically(env_confs, deduplicator)


def compile_topologically(env_confs, deduplicator):
    """Compile environments in topological order of reference."""
    for conf in env_confs:
        env = Environment(in_path=conf['in_path'], deduplicator=deduplicator)
        if env.maybe_create_lockfile():
            # Only munge lockfile if it was written.
            header_text = generate_robust_hash_comment(env.infile) + FEATURES.get_header_text()
            env.replace_header(header_text)
            env.add_references(conf['refs'])



"""Environment discovery"""

import os
import glob
from collections import deque

from toposort import toposort_flatten
from .environment import Environment
from .utils import fix_reference_path, extract_env_name


__all__ = ('discover',)


def discover(glob_pattern):
    """
    Find all files matching given glob_pattern,
    parse them, and return list of environments:

    Recursively follow referenced files not matched by glob_pattern.

    >>> import os
    >>> envs = discover(os.path.join('requirements', '*.in'))
    >>> # import pprint; pprint.pprint(envs)
    >>> envs == [
    ...  {'in_path': os.path.join('requirements', 'base.in'), 'name': 'base',
    ...   'refs': set()},
    ...  {'in_path': os.path.join('requirements', 'test.in'), 'name': 'test',
    ...   'refs': {'base.in'}},
    ...  {'in_path': os.path.join('requirements', 'local.in'), 'name': 'local',
    ...   'refs': {'test.in'}},
    ...  {'in_path': os.path.join('requirements', 'testwin.in'), 'name': 'testwin',
    ...   'refs': {'test.in'}}
    ... ]
    True
    """
    to_visit = deque(map(os.path.normpath, glob.glob(glob_pattern)))
    envs, all_in_paths = {}, set()
    while to_visit:
        in_path = to_visit.pop()
        # name =
        if in_path in all_in_paths:
            continue
        all_in_paths.add(in_path)
        envs[in_path] = {
            'in_path': in_path,
            'name': extract_env_name(in_path),
            'refs': Environment.parse_references(in_path),
        }
        for ref in envs[in_path]['refs']:
            to_visit.append(fix_reference_path(
                orig_path=in_path,
                ref_path=ref
            ))
    return order_by_refs(envs.values())


def order_by_refs(envs):
    """Return topologicaly sorted list of environments.

    I.e. all referenced environments are placed before their references.
    """
    topology = {
        env['in_path']: set(fix_reference_path(env['in_path'], ref)
                            for ref in env['refs'])
        for env in envs
    }
    by_in_path = {
        env['in_path']: env
        for env in envs
    }
    return [
        by_in_path[in_path]
        for in_path in toposort_flatten(topology)
    ]



"""
Live output
===========

Print debug output from pip-compile live.
If the option is disabled (by default) the debug output
is printed only in case of failure.

.. code-block:: text

    --live / --no-live              Print debug output from pip-compile live.
"""
import subprocess

from .base import BaseFeature, ClickOption


class LiveOutput(BaseFeature):
    """Controls whether stdout and stderr should be printed live or at error."""

    OPTION_NAME = 'live'
    CLICK_OPTION = ClickOption(
        long_option='--live/--no-live',
        default=False,
        is_flag=True,
        help_text='Print debug output from pip-compile live.',
    )

    def pipe_arguments(self):
        """Values for stdout and stderr arguments to subprocess.Popen."""
        if self.value:
            return {}
        return {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
        }



"""
Disable upgrades
================

When new dependencies are added it's tempting to keep everything else the same.
To recompile ``.txt`` keeping satisfying version use ``--no-upgrade``:

.. code-block:: text

    --upgrade / --no-upgrade    Upgrade package version (default true)

The option has no effect if there are no existing ``.txt`` files.

Upgrade only selected packages
==============================

To upgrade only one package and keep everything else untouched,
use following option:

.. code-block:: text

    -P, --upgrade-package TEXT  Only upgrade named package.
                                Can be supplied multiple times.

Under the hood it uses `the same option of pip-compile`_
and runs compilation only for files that have one of the passed packages.

This option implies ``--no-upgrade`` and takes precedence over ``--upgrade``.

Thanks to `Jonathan Rogers <https://github.com/JonathanRRogers>`_.

.. _`the same option of pip-compile`: \
        https://github.com/jazzband/pip-tools#updating-requirements
"""

from .base import BaseFeature, ClickOption
from .forward import ForwardOption


class UpgradeAll(ForwardOption):
    """Upgrade all packages in all environments."""

    OPTION_NAME = 'upgrade'
    CLICK_OPTION = ClickOption(
        long_option='--upgrade/--no-upgrade',
        default=True,
        is_flag=True,
        help_text='Upgrade package version (default true)',
    )
    enabled_pin_options = ['--upgrade']

    def __init__(self, controller):
        self._controller = controller

    @property
    def enabled(self):
        """Whether global upgrade is enabled."""
        return self.value and not self._controller.upgrade_selected.active


class UpgradeSelected(BaseFeature):
    """Upgrade only specific packages in all environments."""

    OPTION_NAME = 'upgrade_packages'
    CLICK_OPTION = ClickOption(
        long_option='--upgrade-package',
        short_option='-P',
        multiple=True,
        help_text='Only upgrade named package. '
                  'Can be supplied multiple times.',
    )

    def __init__(self, controller):
        self._controller = controller
        self.reset()

    def reset(self):
        """Clear cached packages."""
        self._env_packages_cache = {}

    @property
    def package_names(self):
        """List of package names to upgrade."""
        return self.value or []

    @property
    def active(self):
        """Whether selective upgrade is active."""
        return bool(self.package_names)

    def pin_options(self):
        """Pin command options for upgrading specific packages."""
        return [
            '--upgrade-package=' + package
            for package in self.package_names
        ]

    def has_package(self, in_path, package_name):
        """Whether specified package name is already in the outfile."""
        return package_name.lower() in self._get_packages(in_path)

    def _get_packages(self, in_path):
        if in_path not in self._env_packages_cache:
            self._env_packages_cache[in_path] = self._read_packages(
                self._compose_output_file_path(in_path)
            )
        return self._env_packages_cache[in_path]

    @staticmethod
    def _read_packages(outfile):
        try:
            with open(outfile, encoding="utf-8") as fp:
                return set(
                    line.split('==', 1)[0].lower()
                    for line in fp
                    if '==' in line
                )
        except IOError:
            # Act as if file is empty
            return set()

    def _compose_output_file_path(self, in_path):
        return self._controller.compose_output_file_path(in_path)

    def affected(self, in_path):
        """Whether environment was affected by upgraded packages."""
        if not self.active:
            return True
        return any(
            self.has_package(in_path, package_name)
            for package_name in self.package_names
        )



"""Dependency class"""

import re

from .features import FEATURES


class Dependency(object):  # pylint: disable=too-many-instance-attributes
    r"""Single dependency.

    Comment may span multiple lines.

    >>> print(Dependency(
    ...   "six==1.0\n "
    ...   "    --hash=abcdef\n"
    ...   "    # via\n"
    ...   "    #   app\n"
    ...   "    #   pkg"
    ... ).serialize())
    six==1.0 \
        --hash=abcdef
        # via
        #   app
        #   pkg
    >>> print(Dependency(
    ...   "six==1.0\n"
    ...   "    # via\n"
    ...   "    #   app\n"
    ...   "    #   pkg"
    ... ).serialize())
    six==1.0
        # via
        #   app
        #   pkg
    >>> # Old-style one-line
    >>> print(Dependency("six==1.0    # via pkg").serialize())
    six==1.0                  # via pkg
    >>> print(Dependency("-e https://site#egg=pkg==1\n   # via lib").serialize())
    https://site#egg=pkg==1
       # via lib
    >>> print(Dependency('dep==1 ; sys_platform == "darwin"').serialize())
    dep==1 ; sys_platform == "darwin"
    """

    COMMENT_JUSTIFICATION = 26

    # Example:
    # unidecode==0.4.21         # via myapp
    # [package]  [version]      [comment]
    RE_DEPENDENCY = re.compile(
        r'(?iu)(?P<package>\S+)'
        r'=='
        r'(?P<version>\S+)'
        r'(?P<markers>\s+;\s+.+?)?'
        r'(?P<hashes>(?:\s*--hash=\S+)+)?'
        r'(?P<comment>(?:\s*#.*)+)?$'
    )
    RE_EDITABLE_FLAG = re.compile(
        r'^-e '
    )
    # -e git+https://github.com/ansible/docutils.git@master#egg=docutils
    # -e "git+https://github.com/zulip/python-zulip-api.git@
    #                 0.4.1#egg=zulip==0.4.1_git&subdirectory=zulip"
    RE_VCS_DEPENDENCY = re.compile(
        r'(?iu)(?P<editable>-e)?'
        r'\s*'
        r'(?P<prefix>\S+#egg=)'
        r'(?P<package>[a-z0-9-_.]+)'
        r'(?P<postfix>\S*)'
        r'(?P<markers>\s+;\s+.+?)?'
        r'(?P<comment>(?:\s*#.*)+)?$'
    )
    # New format, replacing old VCS dependency:
    # docutils @ git+https://github.com/ansible/docutils.git@master
    RE_AT_DEPENDENCY = re.compile(
        r'(?iu)(?P<editable>-e)?'
        r'\s*'
        r'(?P<package>[a-z0-9-_.]+)'
        r' @ '
        r'(?P<url>\S+)'
        r'(?P<markers>\s+;\s+.+?)?'
        r'(?P<comment>(?:\s*#.*)+)?$'
    )
    # Example: 2022.02.1 -> 2022.2.1
    RE_IGNORED_ZEROS = re.compile(r"(?<=\.)0+(?=\d)")

    def __init__(self, line):
        self.is_vcs = False
        self.is_at = False
        self.valid = True
        self.line = line
        vcs = self.RE_VCS_DEPENDENCY.match(line)
        if vcs:
            self.is_vcs = True
            self.package = vcs.group('package')
            self.version = ''
            self.markers = vcs.group('markers') or ''
            self.hashes = ''  # No way!
            self.comment = (vcs.group('comment') or '').rstrip()
            self.comment_span = self._adjust_span(vcs.span('comment'), vcs)
            return
        at_url = self.RE_AT_DEPENDENCY.match(line)
        if at_url:
            self.is_at = True
            self.package = at_url.group('package')
            self.version = ''
            self.markers = at_url.group('markers') or ''
            self.hashes = ''  # ???
            self.comment = (at_url.group('comment') or '').rstrip()
            self.comment_span = self._adjust_span(at_url.span('comment'), at_url)
            return
        regular = self.RE_DEPENDENCY.match(line)
        if regular:
            self.package = regular.group('package')
            self.version = self.RE_IGNORED_ZEROS.sub("", regular.group('version').strip())
            self.markers = regular.group('markers') or ''
            self.hashes = (regular.group('hashes') or '').strip()
            self.comment = (regular.group('comment') or '').rstrip()
            self.comment_span = self._adjust_span(regular.span('comment'), regular)
            return
        self.valid = False

    def serialize(self):
        """
        Render dependency back in string using:
            ~= if package is internal
            == otherwise
        """
        if self.is_vcs or self.is_at:
            return "{}{}".format(
                self.without_editable(self.line[:self.comment_span[0]]).strip(),
                FEATURES.process_dependency_comments(self.comment),
            )
        equal = FEATURES.constraint(self.package)
        package_version = '{package}{equal}{version}  '.format(
            package=self.without_editable(self.package),
            version=self.version,
            equal=equal,
        )
        if self.markers:
            package_version = '{}{}  '.format(package_version.strip(), self.markers)
        if self.hashes:
            hashes = self.hashes.split()
            lines = [package_version.strip()]
            lines.extend(hashes)
            result = ' \\\n    '.join(lines)
            result += FEATURES.process_dependency_comments(self.comment)
            return result
        else:
            if self.comment.startswith('\n'):
                return (
                    package_version.rstrip() +
                    FEATURES.process_dependency_comments(self.comment).rstrip()
                )
            return '{0}{1}'.format(
                package_version.ljust(self.COMMENT_JUSTIFICATION),
                self.comment.lstrip(),
            ).rstrip()  # rstrip for empty comment

    @classmethod
    def without_editable(cls, line):
        """
        Remove the editable flag.
        It's there because pip-compile can't yet do without it
        (see https://github.com/jazzband/pip-tools/issues/272 upstream),
        but in the output of pip-compile it's no longer needed.
        """
        if 'git+git@' in line:
            # git+git can't be installed without -e:
            return line
        return cls.RE_EDITABLE_FLAG.sub('', line)

    def drop_post(self, in_path):
        """Remove .postXXXX postfix from version if needed."""
        self.version = FEATURES.drop_post(in_path, self.package, self.version)

    @staticmethod
    def _adjust_span(span, matchobj):
        if span == (-1, -1):
            length = matchobj.span()[1]
            return (length, length)
        return span



"""
Skip constraints in comments of output files
============================================

When input files contain constraint references (e.g. '-c constraints.in'),
pip-compile adds it to "via" comments. For example::

    rsa==3.4.2
         # via
         #   -c constraints.txt
         #   google-auth

When this option is enabled that snippet will be converted to::

    rsa==3.4.2
         # via google-auth

Saving two lines in .txt file.

.. code-block:: text

    --skip-constraints      Remove constraints from "via" comments.

This feature is especially useful in combination with :ref:`autoresolve`.
"""

import re

from .base import BaseFeature, ClickOption


class SkipConstraintComments(BaseFeature):
    """Remove lines like ``-c file.txt`` from comments in output files."""

    OPTION_NAME = 'skip_constraints'
    CLICK_OPTION = ClickOption(
        long_option='--skip-constraints',
        is_flag=True,
        default=False,
        help_text='Remove constraints from "via" comments.',
    )
    _RE_VIA_COMMENT = re.compile(
        r'^\s*# via$'
    )
    _RE_CONSTRAINT_COMMENT = re.compile(
        r'^\s*#\s+-c \S+$'
    )
    _RE_PACKAGE_COMMENT = re.compile(
        r'^\s*#\s+((?:-r )?\S+)$'
    )

    @property
    def enabled(self):
        """Whether feature was explicitly enabled or not."""
        return self.value

    def process_dependency_comments(self, comment):
        """Remove constraint comments if feature is enabled."""
        if self.enabled:
            return self._drop_sink_comment(comment)
        return comment

    def _drop_sink_comment(self, comment):
        r"""Erase sink constraint from comments.

        >>> feature = SkipConstraintComments()
        >>> feature._drop_sink_comment("\n# via\n#   -c smth\n#   pkg\n")
        '\n# via pkg'
        >>> feature._drop_sink_comment("  # via pkg")
        '  # via pkg'
        """
        lines = comment.splitlines()
        if len(lines) > 2 and self._RE_VIA_COMMENT.match(lines[1]):
            result = lines[:2]
            for line in lines[2:]:
                if self._RE_CONSTRAINT_COMMENT.match(line):
                    continue
                result.append(line)
            return "\n".join(self._collapse_single_via(result))
        return comment

    def _collapse_single_via(self, lines):
        r"""Combine via into a single line when it has only two lines.

        >>> feature = SkipConstraintComments()
        >>> feature._collapse_single_via(["", "# via", "#   pkg"])
        ['', '# via pkg']
        >>> feature._collapse_single_via(["  # via pkg"])
        ['  # via pkg']
        >>> feature._collapse_single_via(["", "# via", "#   -r file"])
        ['', '# via -r file']
        """
        if len(lines) == 3:
            matchobj = self._RE_PACKAGE_COMMENT.match(lines[2])
            if matchobj:
                package = matchobj.group(1)
                return [lines[0], lines[1] + ' ' + package]
        return lines



"""
.. _emit_trusted_host:

Add trusted host annotation
===========================

Control addition of trusted hosts for index URLs to generated files.
Trusted hosts can be defined in ``pip.conf`` file, input requirements file,
or through an argument to ``pip`` command.
Trusted hosts can have invalid HTTPS certificate, or use unencrypted HTTP protocol.

By default, trusted hosts are saved in the generated files.
Pass ``--no-emit-trusted-host`` to remove it.

If no trusted hosts are defined, this flag doesn't have any effect.

.. code-block:: text

    --emit-trusted-host / --no-emit-trusted-host
                                    Add trusted host to generated files
                                    (default true)

See also: `annotate_index`_ and `extra_index_url`_ options.

See pip-tools issue `#382`_ for more details.

.. _#382: https://github.com/jazzband/pip-tools/issues/382
"""

from .base import ClickOption
from .forward import ForwardOption


class EmitTrustedHost(ForwardOption):
    """Optionally add the trusted host to the generated files."""

    OPTION_NAME = 'emit_trusted_host'
    CLICK_OPTION = ClickOption(
        long_option='--emit-trusted-host/--no-emit-trusted-host',
        is_flag=True,
        default=True,
        help_text="Add trusted host option to generated file"
    )
    enabled_pin_options = ['--emit-trusted-host']
    disabled_pin_options = ['--no-emit-trusted-host']



"""
Requirements Directory
======================

While it's a common practice to put requirements files inside ``requirements``
directory, it's not always the case.
The directory can be overridden with this option:

.. code-block:: text

    -d, --directory TEXT   Directory path with requirements files
"""

import os

from .base import BaseFeature, ClickOption


class BaseDir(BaseFeature):
    """Override input file extension."""

    OPTION_NAME = 'base_dir'
    CLICK_OPTION = ClickOption(
        long_option='--directory',
        short_option='-d',
        default="requirements",
        is_flag=False,
        help_text='Directory path with requirements files.',
    )

    @property
    def path(self):
        """Get the base directory path.

        >>> BaseDir().path == 'requirements'
        True
        """
        return self.value

    def file_path(self, file_name):
        """Compose file path for a given file name.

        >>> import os.path
        >>> expected = os.path.join('requirements', 'base.txt')
        >>> expected == BaseDir().file_path('base.txt')
        True
        """
        return os.path.join(self.value, file_name)



"""
Forbid .postX release
=====================

``pip-compile-multi`` can remove ``.postX`` part of dependencies versions.

.. code-block:: text

    -p, --forbid-post TEXT      Environment name (base, test, etc) that cannot
                                have packages with post-release versions
                                (1.2.3.post777).
                                Can be supplied multiple times.

.. note::

    Be careful with this option since different maintainers treat
    post releases differently.
"""

from .base import BaseFeature, ClickOption


class ForbidPost(BaseFeature):
    """Truncate postXXX from versions for selected packages."""

    OPTION_NAME = 'forbid_post'
    CLICK_OPTION = ClickOption(
        long_option='--forbid-post',
        short_option='-p',
        multiple=True,
        help_text="Environment name (base, test, etc) that cannot have "
                  'packages with post-release versions (1.2.3.post777). '
                  'Can be supplied multiple times.'
    )

    @property
    def enabled_envs(self):
        """Convert to set."""
        return set(self.value or [])

    @staticmethod
    def drop_post(version):
        """Remove .postXXXX postfix from version.

        >>> ForbidPost.drop_post('1.2.3.post123')
        '1.2.3'
        >>> ForbidPost.drop_post('1.2.3')
        '1.2.3'
        """
        post_index = version.find('.post')
        if post_index >= 0:
            return version[:post_index]
        return version

    def post_forbidden(self, env_name):
        """Whether post versions are forbidden for passed environment name."""
        return env_name in self.enabled_envs



"""Remove packages included in referenced environments."""

import logging

from pipcompilemulti.utils import recursive_refs, merged_packages


logger = logging.getLogger("pip-compile-multi")


class PackageDeduplicator:
    """Remove packages included in referenced environments."""

    def __init__(self):
        self.env_packages = {}
        self.env_confs = None

    def on_discover(self, env_confs):
        """Save environment references."""
        self.env_confs = env_confs

    def register_packages_for_env(self, in_path, packages):
        """Save environment packages."""
        self.env_packages[in_path] = packages

    def ignored_packages(self, in_path):
        """Get package mapping from name to version for referenced environments."""
        if self.env_confs is None:
            return {}
        rrefs = recursive_refs(self.env_confs, in_path)
        return IgnoredPackages(merged_packages(self.env_packages, rrefs))

    def recursive_refs(self, in_path):
        """Return recursive list of environment names referenced by in_path."""
        if self.env_confs is None:
            return {}
        return recursive_refs(self.env_confs, in_path)


class IgnoredPackages:
    """Mapping from package name to version.

    Handles name normalization for packages like:
    zope.interface, zope-interface, zope_interface.
    """
    _DELIMITERS = ('_', '-', '.')

    def __init__(self, package_versions):
        self._package_versions = package_versions
        self._stems = {
            self._make_stem(name): name
            for name in self._package_versions
        }

    def __getitem__(self, key):
        canonical_key = self._stems[self._make_stem(key)]
        return self._package_versions[canonical_key]

    def __contains__(self, key):
        return self._make_stem(key) in self._stems

    @classmethod
    def _make_stem(cls, name):
        for delim in cls._DELIMITERS:
            name = name.replace(delim, '-')
        return name.lower()



"""Get tasks options from INI file"""
import sys
import configparser

from .options import LIST_OPTIONS


def read_config():
    """Read requirements.ini and return list of pairs (name, options)
    If no requirements sections found, return None.
    If some sections found, but none matches current runtime, return empty list.
    """
    return filter_sections(read_sections())


def filter_sections(sections):
    """Filter through pairs (name, options)
    leaving only those that match runtime.

    If no requirements sections found, return None.
    If some sections found, but none matches current runtime, return empty list.
    """
    if not sections:
        return None
    jobs = []
    matchers = python_version_matchers()
    for name, options in sections:
        target_version = options.pop('python', None)
        if target_version in matchers:
            jobs.append((name, options))
    return jobs


def read_sections():
    """Read ini files and return list of pairs (name, options)"""
    config = configparser.ConfigParser()
    config.read(('requirements.ini', 'setup.cfg', 'tox.ini'))
    return [
        (
            name,
            {
                key: parse_value(key, config[name][key])
                for key in config[name]
            }
        )
        for name in config.sections()
        if 'requirements' in name
    ]


def parse_value(key, value):
    """Parse value as comma-delimited list if key is in LIST_OPTIONS"""
    if key in LIST_OPTIONS:
        return [item.strip()
                for item in value.split(',')]
    return value


def python_version_matchers():
    """Return set of string representations of current python version"""
    version = sys.version_info
    patterns = [
        "{0}",
        "{0}{1}",
        "{0}.{1}",
    ]
    matchers = [
        pattern.format(*version)
        for pattern in patterns
    ] + [None]
    return set(matchers)



"""
Build isolation
===============

Allows disabling build isolation through the equivalent ``pip-compile`` flag.
Build isolation is enabled by default.

.. code-block:: text

  --build-isolation / --no-build-isolation
                                  Enable isolation when building a modern
                                  source distribution. Build dependencies
                                  specified by PEP 518 must be already
                                  installed if build isolation is disabled.
"""
from .base import ClickOption
from .forward import ForwardOption


class BuildIsolation(ForwardOption):
    """Proxies build isolation flag to pip-compile"""

    OPTION_NAME = 'build_isolation'
    CLICK_OPTION = ClickOption(
        long_option='--build-isolation/--no-build-isolation',
        is_flag=True,
        default=True,
        help_text=(
            'Enable isolation when building a modern source distribution. '
            'Build dependencies specified by PEP 518 must be already '
            'installed if build isolation is disabled.'
        ),
    )
    disabled_pin_options = ['--no-build-isolation']



"""
Generate hashes
===============

Put package hash after pinned version for additional security.
Format for this option is

.. code-block:: text

  -g, --generate-hashes TEXT  Input file name (base.in, requirements/test.in, etc)
                              that needs packages hashes.
                              Can be supplied multiple times.
                              For backwards compatibility can be short
                              environment name (base, test, etc.)

Example invocation:

.. code-block:: shell

    $ pip-compile-multi -g base -g docs

Example output:

.. code-block:: text

    pip-tools==1.11.0 \
        --hash=sha256:50288eb066ce66dbef5401a21530712a93c659fe480c7d8d34e2379300555fa1 \
        --hash=sha256:ba427b68443466c389e3b0b0ef55f537ab39344190ea980dfebb333d0e6a50a3
    first==2.0.1 \
        --hash=sha256:3bb3de3582cb27071cfb514f00ed784dc444b7f96dc21e140de65fe00585c95e \
        --hash=sha256:41d5b64e70507d0c3ca742d68010a76060eea8a3d863e9b5130ab11a4a91aa0e \
        # via pip-tools

``pip`` requires all packages to have hashes if at least one has it.
``pip-compile-multi`` will recursively propagate this option to all
environments that are referencing or referenced by selected environments.
"""  # noqa: E501
import os

from pipcompilemulti.utils import reference_cluster
from .base import BaseFeature, ClickOption


class AddHashes(BaseFeature):
    """Write hashes for pinned packages."""

    OPTION_NAME = 'add_hashes'
    CLICK_OPTION = ClickOption(
        long_option='--generate-hashes',
        short_option='-g',
        multiple=True,
        help_text='Input file name (base.in, requirements/test.in, etc) '
                  'that needs packages hashes. '
                  'Can be supplied multiple times.',

    )

    def __init__(self, controller):
        self._controller = controller
        self._hashed_by_reference = None

    @property
    def enabled_in_paths(self):
        """Convert list of .in paths to a set.

        For backwards compatibility, check if passed value is env name
        and convert it to in_path.
        """
        names_or_paths = self.value or []
        in_paths = set()
        for name_or_path in names_or_paths:
            in_path = self._controller.compose_input_file_path(name_or_path)
            if os.path.exists(in_path):
                in_paths.add(in_path)
            else:
                in_paths.add(name_or_path)
        return in_paths

    def on_discover(self, env_confs):
        """Save environment names that need hashing."""
        self._hashed_by_reference = set()
        for in_path in self.enabled_in_paths:
            self._hashed_by_reference.update(
                reference_cluster(env_confs, in_path)
            )

    def _needs_hashes(self, in_path):
        assert self._hashed_by_reference is not None
        return in_path in self._hashed_by_reference

    def pin_options(self, in_path):
        """Return --generate-hashes if env requires it."""
        if self._needs_hashes(in_path):
            return ['--generate-hashes']
        return []



"""Package configuration"""

import os
from setuptools import setup, find_packages


VERSION = "2.6.2"


README = """
pip-compile-multi
=================

Compile multiple requirements files to lock dependency versions.

Install
-------

.. code-block:: shell

    pip install pip-compile-multi

Run
----

.. code-block:: shell

    pip-compile-multi


Links
-----

* Documentation: https://pip-compile-multi.readthedocs.io/en/latest/
* Releases: https://pypi.python.org/pypi/pip-compile-multi
* Code: https://github.com/peterdemin/pip-compile-multi
* Issue tracker: https://github.com/peterdemin/pip-compile-multi/issues

"""


with open('HISTORY.rst', encoding='utf-8') as fp:
    HISTORY = fp.read().replace('.. :changelog:', '')


with open(os.path.join('requirements', 'base.in'), encoding='utf-8') as fp:
    REQUIREMENTS = list(fp)


CONSOLE_SCRIPTS = [
    'pip-compile-multi = pipcompilemulti.cli_v1:cli',
]
if os.environ.get('PCM_ALPHA') == 'ON':
    CONSOLE_SCRIPTS.append(
        'requirements = pipcompilemulti.cli_v2:cli'
    )


setup(
    name='pip-compile-multi',
    version=VERSION,
    description="Compile multiple requirements files "
                "to lock dependency versions",
    long_description=README + '\n\n' + HISTORY,
    author='Peter Demin',
    author_email='peterdemin@gmail.com',
    url='https://github.com/peterdemin/pip-compile-multi',
    include_package_data=True,
    packages=find_packages(exclude=['tests']),
    install_requires=REQUIREMENTS,
    python_requires='~=3.6',
    license="MIT",
    zip_safe=False,
    keywords='pip-compile-multi',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    entry_points={
        'console_scripts': CONSOLE_SCRIPTS,
    },
    setup_requires=['setuptools', 'wheel'],
)



"""
.. _limit-in-files:

Limit input files
=================

By default ``pip-compile-multi`` compiles all ``.in`` files in ``requirements`` directory.
To limit compilation to only a subset, use

.. code-block:: text

    -t, --only-path TEXT        Compile only for passed input file paths and
                                their references.
                                Can be supplied multiple times.

For example, to compile one file under Python2.7 and another under Python3.6, run:

.. code-block:: text

    $ virtual-env27/bin/pip-compile-multi -t requirements/deps27.in
    Locking requirements/deps27.in to requirements/deps27.txt. References: []
    $ virtual-env36/bin/pip-compile-multi -t requirements/deps36.in
    Locking requirements/deps36.in to requirements/deps36.txt. References: []
"""

from pipcompilemulti.utils import recursive_refs
from .base import BaseFeature, ClickOption


class LimitInPaths(BaseFeature):
    """Limit discovered input files to specified subset.

    >>> from pipcompilemulti.options import OPTIONS
    >>> feature = LimitInPaths()
    >>> OPTIONS[feature.OPTION_NAME] = ['test.in']
    >>> feature.on_discover([
    ...     {'in_path': 'base.in', 'refs': []},
    ...     {'in_path': 'test.in', 'refs': ['base.in']},
    ...     {'in_path': 'docs.in', 'refs': []},
    ... ])
    >>> feature.included('base.in')
    True
    >>> feature.included('test.in')
    True
    >>> feature.included('docs.in')
    False
    """

    OPTION_NAME = 'include_in_paths'
    CLICK_OPTION = ClickOption(
        long_option='--only-path',
        short_option='-t',
        multiple=True,
        help_text='Compile only for passed input paths and their '
                  'references. Can be supplied multiple times.',
    )

    def __init__(self):
        self._all_envs = None

    @property
    def direct_envs(self):
        """Set of environments included by command line options."""
        return set(self.value or [])

    def on_discover(self, env_confs):
        """Save set of all (recursive) included environments."""
        if not self.direct_envs:
            # No limit means all envs included:
            self._all_envs = [env['in_path'] for env in env_confs]
            return
        transitive_refs = {
            ref
            for in_path in self.direct_envs
            for ref in recursive_refs(env_confs, in_path)
        }
        self._all_envs = self.direct_envs | transitive_refs

    def included(self, in_path):
        """Whether environment is included directly or by reference."""
        return in_path in self._all_envs



"""
Custom Header
=============

``pip-compile-multi`` adds a brief header into generated files.
Override it with

.. code-block:: text

    -h, --header TEXT      File path with custom header text for generated files
"""

from .base import BaseFeature, ClickOption


DEFAULT_HEADER = """
#
# This file is autogenerated by pip-compile-multi
# To update, run:
#
#    pip-compile-multi
#
""".lstrip()


class CustomHeader(BaseFeature):
    """Put custom header at the beginning of locked files."""

    OPTION_NAME = 'header_file'
    CLICK_OPTION = ClickOption(
        long_option='--header',
        short_option='-h',
        default='',
        help_text='File path with custom header text for generated files.',
    )

    def __init__(self):
        self._header_text = None

    @property
    def text(self):
        """Text to put in the beginning of each generated file."""
        if self._header_text is None:
            if self.value:
                self._header_text = self._read_header_text()
            else:
                self._header_text = DEFAULT_HEADER
        return self._header_text

    def _read_header_text(self):
        with open(self.value, encoding="utf-8") as fp:
            return fp.read()



"""Functional utilities for lists and dicts manipulation."""

import os
import logging
import itertools


logger = logging.getLogger("pip-compile-multi")


def extract_env_name(file_path):
    """Return environment name for given requirements file path.

    >>> extract_env_name("base.in")
    'base'
    >>> extract_env_name("sub/req.in")
    'req'
    """
    return os.path.splitext(os.path.basename(file_path))[0]


def fix_reference_path(orig_path, ref_path):
    """Find actual path to reference using relative path to original file.

    >>> fix_reference_path("dir/file", "../ref")
    'ref'
    """
    return os.path.normpath(os.path.join(os.path.dirname(orig_path), ref_path))


def recursive_refs(envs, in_path):
    """Return set of recursive refs for given env name."""
    refs_by_in_path = {
        env['in_path']: {
            fix_reference_path(env['in_path'], ref)
            for ref in env['refs']
        }
        for env in envs
    }
    refs = refs_by_in_path[in_path]
    if refs:
        indirect_refs = set(
            subref
            for ref in refs
            for subref in recursive_refs(envs, ref)
        )
    else:
        indirect_refs = set()
    return set.union(refs, indirect_refs)


def merged_packages(env_packages, names):
    """Return union set of environment packages with given names.

    >>> sorted(merged_packages(
    ...     {
    ...         'a': {'x': 1, 'y': 2},
    ...         'b': {'y': 2, 'z': 3},
    ...         'c': {'z': 3, 'w': 4}
    ...     },
    ...     ['a', 'b']
    ... ).items())
    [('x', 1), ('y', 2), ('z', 3)]
    """
    combined_packages = sorted(itertools.chain.from_iterable(
        env_packages[name].items()
        for name in names
    ))
    result = {}
    errors = set()
    for name, version in combined_packages:
        if name in result:
            if result[name] != version:
                errors.add((name, version, result[name]))
        else:
            result[name] = version
    if errors:
        for error in sorted(errors):
            logger.error(
                "Package %s was resolved to different "
                "versions in different environments: %s and %s",
                error[0], error[1], error[2],
            )
        raise RuntimeError(
            "Please add constraints for the package version listed above"
        )
    return result


def reference_cluster(envs, in_path):
    """
    Return set of all env in_paths referencing or
    referenced by given in_path.

    >>> cluster = sorted(reference_cluster([
    ...     {'in_path': 'base', 'refs': []},
    ...     {'in_path': 'test', 'refs': ['base']},
    ...     {'in_path': 'local', 'refs': ['test']},
    ... ], 'test'))
    >>> cluster == ['base', 'local', 'test']
    True
    """
    edges = [
        set([env['in_path'], fix_reference_path(env['in_path'], ref)])
        for env in envs
        for ref in env['refs']
    ]
    prev, cluster = set(), set([in_path])
    while prev != cluster:
        # While cluster grows
        prev = set(cluster)
        to_visit = []
        for edge in edges:
            if cluster & edge:
                # Add adjacent nodes:
                cluster |= edge
            else:
                # Leave only edges that are out
                # of cluster for the next round:
                to_visit.append(edge)
        edges = to_visit
    return cluster



"""
Allow Unsafe Packages
=====================

If your project depends on packages that include ``setuptools``
or other packages considered "unsafe" by ``pip-tools`` and you
still wish to have them included in the resulting requirements files,
you can pass this option to do so:

.. code-block:: text

    -s, --allow-unsafe          Whether or not to include 'unsafe' packages
                                in generated requirements files. Consult
                                pip-compile --help for more information

This is commonly used with --generate-hashes to avoid
generating requirements files which `cannot be installed`_.

.. _cannot be installed: https://github.com/jazzband/pip-tools/issues/806
"""

from .base import ClickOption
from .forward import ForwardOption


class AllowUnsafe(ForwardOption):
    """Use pip-tools cache, or rebuild from scratch."""

    OPTION_NAME = 'allow_unsafe'
    CLICK_OPTION = ClickOption(
        long_option='--allow-unsafe',
        short_option='-s',
        is_flag=True,
        default=False,
        help_text="Whether or not to include 'unsafe' packages "
                  'in generated requirements files. '
                  'Consult pip-compile --help for more information'
    )
    enabled_pin_options = ['--allow-unsafe']



"""Base feature for forwarding pip-tools options."""

from .base import BaseFeature


class ForwardOption(BaseFeature):
    """Forward command line option to pip-tools."""

    #: Pin command options when feature is enabled.
    enabled_pin_options = []
    #: Pin command options when feature is disabled.
    disabled_pin_options = []

    @property
    def enabled(self):
        """Is feature enabled."""
        return self.value

    def pin_options(self):
        """Pin command options."""
        if self.enabled:
            return self.enabled_pin_options
        return self.disabled_pin_options



"""
Check that ``pip-compile-multi`` was run after changes in ``.in`` file
======================================================================

``pip-compile-multi`` adds a special line (before header) at the beginning of each generated file.
This line contains a SHA1 hash of the ``.in`` file's contents.

Command

.. code-block:: shell

    $ pip-compile-multi verify
    Verifying that requirements/base.txt was generated from requirements/base.in.
    Success - comments match.
    Verifying that requirements/test.txt was generated from requirements/test.in.
    Success - comments match.
    Verifying that requirements/local.txt was generated from requirements/local.in.
    Success - comments match.

recalculates hashes for ``.in`` files and compares them with the stored values.

If verification fails, an error message is logged and exit code 1 is returned:

.. code-block:: shell

    $ pip-compile-multi verify
    Verifying that requirements/base.txt was generated from requirements/base.in.
    Success - comments match.
    Verifying that requirements/test.txt was generated from requirements/test.in.
    FAILURE!
    Expecting: # SHA1:c93d71964e14b04f3c8327d16dbc4d6b1bbc3b1d
    Found:     # SHA1:6c2562322ca1bdc8309b08581a2aa4efbb5a4534
    Verifying that requirements/local.txt was generated from requirements/local.in.
    Success - comments match.


In big teams it might be a good idea to have this check in ``tox.ini``:

.. code-block:: ini

    [testenv:verify]
    skipsdist = true
    skip_install = true
    deps = pip-compile-multi
    commands = pip-compile-multi verify
    whitelist_externals = pip-compile-multi
"""

import hashlib
import logging

from .discover import discover
from .environment import Environment
from .features import FEATURES


logger = logging.getLogger("pip-compile-multi")


def verify_environments():
    """
    For each environment verify hash comments and report failures.
    If any failure occured, exit with code 1.
    """
    env_confs = discover(FEATURES.compose_input_file_path('*'))
    success = True
    for conf in env_confs:
        env = Environment(in_path=conf['in_path'])
        current_comment = generate_hash_comment(env.infile)
        robust_comment = generate_robust_hash_comment(env.infile)
        existing_comment = parse_hash_comment(env.outfile)
        if existing_comment in (robust_comment, current_comment):
            logger.info("OK - %s was generated from %s.",
                        env.outfile, env.infile)
        else:
            logger.error("ERROR! %s was not regenerated after changes in %s.",
                         env.outfile, env.infile)
            logger.error("Expecting: %s", robust_comment.strip())
            logger.error("Found:     %s", existing_comment.strip())
            success = False
    return success


def generate_hash_comment(file_path):
    """
    Read file with given file_path and return string of format

        # SHA1:da39a3ee5e6b4b0d3255bfef95601890afd80709

    which is hex representation of SHA1 file content hash
    """
    with open(file_path, 'rb') as fp:
        hexdigest = hashlib.sha1(fp.read().strip()).hexdigest()
    return f"# SHA1:{hexdigest}\n"


def generate_robust_hash_comment(file_path):
    """
    Read file with given file_path and return string of format

        # SHA1:da39a3ee5e6b4b0d3255bfef95601890afd80709

    which is hex representation of SHA1 file content hash.
    File content is pre-processed by stripping comments, whitespace and newlines.
    """
    with open(file_path, 'rt', encoding="utf-8") as fp:
        essense = ''.join(sorted(
            line.split('#')[0].strip()
            for line in fp
        ))
    hexdigest = hashlib.sha1(essense.encode("utf-8")).hexdigest()
    return f"# SHA1:{hexdigest}\n"


def parse_hash_comment(file_path):
    """
    Read file with given file_path line by line,
    return the first line that starts with "# SHA1:", like this:

        # SHA1:da39a3ee5e6b4b0d3255bfef95601890afd80709
    """
    with open(file_path, encoding="utf-8") as fp:
        for line in fp:
            if line.startswith("# SHA1:"):
                return line
    return ''
